#include <linux/init.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/sched/signal.h>
#include <linux/pid.h>

MODULE_LICENSE("Dual BSD/GPL");

void ps_get_pid(struct task_struct *task, char *result)
{
	sprintf(result, "%d(%d)\t", task->pid, task->tgid);
}

void ps_get_command(struct task_struct *task, char *result)
{
	sprintf(result, "%-16s\t", task->comm);
}

void ps_get_state(struct task_struct *task, char *result)
{
	if (task->__state == TASK_RUNNING) sprintf(result, "R(%x)\t", task->__state);
	else if ((task->__state & TASK_INTERRUPTIBLE) == TASK_INTERRUPTIBLE) sprintf(result, "S(%x)\t", task->__state);
	else if ((task->__state & TASK_UNINTERRUPTIBLE) == TASK_UNINTERRUPTIBLE) sprintf(result, "I(%x)\t", task->__state);
	else sprintf(result, "U\t");
	
}

void ps_get_uid(struct task_struct *task, char *result)
{
	// place your implementation here
}

void ps_get_ppid(struct task_struct *task, char *result)
{
	// place your implementation here
}

void ps_get_nice(struct task_struct *task, char *result)
{
	// place your implementation here
}

void ps_get_rtprio(struct task_struct *task, char *result)
{
	// place your implementation here
}

void ps_get_priority(struct task_struct *task, char *result)
{
	// place your implementation here
}

void ps_get_time(struct task_struct *task, char *result)
{
	// place your implementation here
}

void ps_get_schedpolicy(struct task_struct *task, char *result)
{
	// place your implementation here
}

static void show_info(struct task_struct *task)
{
	char result[256];
	
	memset(result, 0, sizeof(result));
	ps_get_pid(task, result + strlen(result));
	ps_get_command(task, result + strlen(result));
	ps_get_state(task, result + strlen(result));
	ps_get_uid(task, result + strlen(result));
	ps_get_ppid(task, result + strlen(result));
	ps_get_nice(task, result + strlen(result));
	ps_get_rtprio(task, result + strlen(result));
	ps_get_priority(task, result + strlen(result));
	ps_get_time(task, result + strlen(result));		
	ps_get_schedpolicy(task, result + strlen(result));
	printk("%s", result);		
}

static int ps_init(void) 
{
	struct task_struct *process, *thread;

	
	printk(KERN_ALERT "ps enter: %d\n", init_task.pid);
	
	printk("PID\tCOMMAND\t\t\tSTATE\tUID\tPPID\tNI\tRTPRIO\tPRI\tTIME\t\tSCH\t");

	// method-1: iterate through the thread_node linked list 
	//           whose head is in the signal_struct shared by the thread group
	for_each_process_thread(process, thread) 
	{	
		show_info(thread);	
	}

	// method-2: iterate through the thread_group linked list 
	//           whose head is in the task_struct of the thread group leader
	for_each_process(process)
	{
		show_info(process);
		list_for_each_entry(thread, &(process->thread_group), thread_group)
		{
			show_info(thread);		
		}
	}
	
	return 0;
}

static void ps_exit(void) 
{
	printk(KERN_ALERT "ps exit\n");
}

module_init(ps_init);
module_exit(ps_exit);

MODULE_AUTHOR("UV");
MODULE_DESCRIPTION("A simple module to list detailed information of procsses");
